@@ -7,7 +7,8 @@ from .product import *
 from .stock import *
 from .configuration import *
 from .invoice import *
-from .party import PartyReplace
+from . import party
+from . import sale_reporting
 
 
 def register():
@@ -30,6 +31,19 @@ def register():
         Configuration,
         ConfigurationSequence,
         ConfigurationSaleMethod,
+        sale_reporting.Context,
+        sale_reporting.Customer,
+        sale_reporting.CustomerTimeseries,
+        sale_reporting.Product,
+        sale_reporting.ProductTimeseries,
+        sale_reporting.Category,
+        sale_reporting.CategoryTimeseries,
+        sale_reporting.CategoryTree,
+        sale_reporting.Country,
+        sale_reporting.CountryTimeseries,
+        sale_reporting.Subdivision,
+        sale_reporting.SubdivisionTimeseries,
+        sale_reporting.Region,
         Invoice,
         InvoiceLine,
         module='sale', type_='model')
@@ -38,7 +52,10 @@ def register():
         HandleShipmentException,
         HandleInvoiceException,
         ReturnSale,
-        PartyReplace,
+        ModifyHeader,
+        party.PartyReplace,
+        party.PartyErase,
+        sale_reporting.OpenRegion,
         module='sale', type_='wizard')
     Pool.register(
         SaleReport,
