@@ -5,17 +5,12 @@ from decimal import Decimal
 from itertools import groupby, chain
 from functools import partial
 
-from sql import Table, Null
-from sql.functions import Overlay, Position
-from sql.operators import Concat
-
-from trytond.model import Workflow, ModelView, ModelSQL, fields, \
+from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, \
     sequence_ordered
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
     Button
-from trytond import backend
-from trytond.pyson import If, Eval, Bool, PYSONEncoder, Id
+from trytond.pyson import If, Eval, Bool, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
@@ -27,7 +22,8 @@ __all__ = ['Sale', 'SaleIgnoredInvoice', 'SaleRecreatedInvoice',
     'SaleLineRecreatedMove', 'SaleReport', 'OpenCustomer',
     'HandleShipmentExceptionAsk', 'HandleShipmentException',
     'HandleInvoiceExceptionAsk', 'HandleInvoiceException',
-    'ReturnSaleStart', 'ReturnSale']
+    'ReturnSaleStart', 'ReturnSale', 'ModifyHeader',
+    'get_shipments_returns', 'search_shipments_returns']
 
 _ZERO = Decimal(0)
 STATES = [
@@ -40,6 +36,36 @@ STATES = [
     ]
 
 
+def get_shipments_returns(model_name):
+    "Computes the returns or shipments"
+    def method(self, name):
+        Model = Pool().get(model_name)
+        shipments = set()
+        for line in self.lines:
+            for move in line.moves:
+                if isinstance(move.shipment, Model):
+                    shipments.add(move.shipment.id)
+        return list(shipments)
+    return method
+
+
+def search_shipments_returns(model_name):
+    "Search on shipments or returns"
+    def method(self, name, clause):
+        nested = clause[0].lstrip(name)
+        if nested:
+            return [('lines.moves.shipment' + nested,)
+                + tuple(clause[1:3]) + (model_name,) + tuple(clause[3:])]
+        else:
+            if isinstance(clause[2], str):
+                target = 'rec_name'
+            else:
+                target = 'id'
+            return [('lines.moves.shipment.' + target,)
+                + tuple(clause[1:3]) + (model_name,)]
+    return classmethod(method)
+
+
 class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     'Sale'
     __name__ = 'sale.sale'
@@ -211,6 +237,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                 ('draft', 'quotation'),
                 ('quotation', 'confirmed'),
                 ('confirmed', 'processing'),
+                ('confirmed', 'draft'),
                 ('processing', 'processing'),
                 ('processing', 'done'),
                 ('done', 'processing'),
@@ -222,33 +249,43 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         cls._buttons.update({
                 'cancel': {
                     'invisible': ~Eval('state').in_(['draft', 'quotation']),
+                    'depends': ['state'],
                     },
                 'draft': {
-                    'invisible': ~Eval('state').in_(['cancel', 'quotation']),
-                    'icon': If(Eval('state') == 'cancel', 'tryton-clear',
-                        'tryton-go-previous'),
+                    'invisible': ~Eval('state').in_(
+                        ['cancel', 'quotation', 'confirmed']),
+                    'icon': If(Eval('state') == 'cancel',
+                        'tryton-undo',
+                        'tryton-back'),
+                    'depends': ['state'],
                     },
                 'quote': {
                     'invisible': Eval('state') != 'draft',
                     'readonly': ~Eval('lines', []),
+                    'depends': ['state'],
                     },
                 'confirm': {
                     'invisible': Eval('state') != 'quotation',
+                    'depends': ['state'],
                     },
                 'process': {
                     'invisible': Eval('state') != 'confirmed',
+                    'depends': ['state'],
                     },
                 'handle_invoice_exception': {
                     'invisible': ((Eval('invoice_state') != 'exception')
                         | (Eval('state') == 'cancel')),
-                    'readonly': ~Eval('groups', []).contains(
-                        Id('sale', 'group_sale')),
+                    'depends': ['state', 'invoice_state'],
                     },
                 'handle_shipment_exception': {
                     'invisible': ((Eval('shipment_state') != 'exception')
                         | (Eval('state') == 'cancel')),
-                    'readonly': ~Eval('groups', []).contains(
-                        Id('sale', 'group_sale')),
+                    'depends': ['state', 'shipment_state'],
+                    },
+                'modify_header': {
+                    'invisible': ((Eval('state') != 'draft')
+                        | ~Eval('lines', [-1])),
+                    'depends': ['state'],
                     },
                 })
         # The states where amounts are cached
@@ -256,42 +293,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
 
     @classmethod
     def __register__(cls, module_name):
-        pool = Pool()
-        SaleLine = pool.get('sale.line')
-        TableHandler = backend.get('TableHandler')
-        sale_line_invoice_line_table_name = 'sale_line_invoice_lines_rel'
-        Move = pool.get('stock.move')
-        cursor = Transaction().connection.cursor()
-        model_data = Table('ir_model_data')
-        model_field = Table('ir_model_field')
-        sql_table = cls.__table__()
-
-        # Migration from 1.2: packing renamed into shipment
-        cursor.execute(*model_data.update(
-                columns=[model_data.fs_id],
-                values=[Overlay(model_data.fs_id, 'shipment',
-                        Position('packing', model_data.fs_id),
-                        len('packing'))],
-                where=model_data.fs_id.like('%packing%')
-                & (model_data.module == module_name)))
-        cursor.execute(*model_field.update(
-                columns=[model_field.relation],
-                values=[Overlay(model_field.relation, 'shipment',
-                        Position('packing', model_field.relation),
-                        len('packing'))],
-                where=model_field.relation.like('%packing%')
-                & (model_field.module == module_name)))
-        cursor.execute(*model_field.update(
-                columns=[model_field.name],
-                values=[Overlay(model_field.name, 'shipment',
-                        Position('packing', model_field.name),
-                        len('packing'))],
-                where=model_field.name.like('%packing%')
-                & (model_field.module == module_name)))
-        table = TableHandler(cls, module_name)
-        table.column_rename('packing_state', 'shipment_state')
-        table.column_rename('packing_method', 'shipment_method')
-        table.column_rename('packing_address', 'shipment_address')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8: rename reference into number
         if (table.column_exist('reference')
@@ -299,52 +301,14 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             table.column_rename('reference', 'number')
 
         super(Sale, cls).__register__(module_name)
-
-        # Migration from 1.2
-        cursor.execute(*sql_table.update(
-                columns=[sql_table.invoice_method],
-                values=['shipment'],
-                where=sql_table.invoice_method == 'packing'))
-
-        table = TableHandler(cls, module_name)
-        # Migration from 2.2
-        table.not_null_action('sale_date', 'remove')
-
-        # state confirmed splitted into confirmed and processing
-        if (TableHandler.table_exist(SaleLine._table)
-                and TableHandler.table_exist(
-                    sale_line_invoice_line_table_name)
-                and TableHandler.table_exist(Move._table)):
-            sale_line = SaleLine.__table__()
-            sale_line_invoice_line = \
-                Table(sale_line_invoice_line_table_name)
-            move = Move.__table__()
-            # Wrap subquery inside an other inner subquery because MySQL syntax
-            # doesn't allow update a table and select from the same table in a
-            # subquery.
-            sub_query = sql_table.join(sale_line,
-                condition=sale_line.sale == sql_table.id
-                ).join(sale_line_invoice_line, 'LEFT',
-                    condition=sale_line_invoice_line.sale_line == sale_line.id
-                    ).join(move, 'LEFT',
-                        condition=(move.origin == Concat(
-                                SaleLine.__name__ + ',', sale_line.id))
-                        ).select(sql_table.id,
-                            where=(sql_table.state == 'confirmed')
-                            & ((sale_line_invoice_line.id != Null)
-                                | (move.id != Null)))
-            cursor.execute(*sql_table.update(
-                    columns=[sql_table.state],
-                    values=['processing'],
-                    where=sql_table.id.in_(sub_query.select(sub_query.id))))
-
-        # Add index on create_date
-        table = TableHandler(cls, module_name)
-        table.index_action('create_date', action='add')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 4.0: Drop not null on payment_term
         table.not_null_action('payment_term', 'remove')
 
+        # Add index on create_date
+        table.index_action('create_date', action='add')
+
     @classmethod
     def default_payment_term(cls):
         PaymentTerm = Pool().get('account.invoice.payment_term')
@@ -453,7 +417,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             for line in self.lines:
                 self.untaxed_amount += getattr(line, 'amount', None) or 0
             taxes = self._get_taxes()
-            self.tax_amount = sum(v['amount'] for v in taxes.itervalues())
+            self.tax_amount = sum(v['amount'] for v in taxes.values())
         if self.currency:
             self.untaxed_amount = self.currency.round(self.untaxed_amount)
             self.tax_amount = self.currency.round(self.tax_amount)
@@ -481,7 +445,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
 
     def get_tax_amount(self):
         return sum(
-            (v['amount'] for v in self._get_taxes().itervalues()), Decimal(0))
+            (v['amount'] for v in self._get_taxes().values()), Decimal(0))
 
     @classmethod
     def get_amount(cls, sales, names):
@@ -520,7 +484,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             'tax_amount': tax_amount,
             'total_amount': total_amount,
             }
-        for key in result.keys():
+        for key in list(result.keys()):
             if key not in names:
                 del result[key]
         return result
@@ -564,39 +528,9 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                     'invoice_state': state,
                     })
 
-    def get_shipments_returns(model_name):
-        "Computes the returns or shipments"
-        def method(self, name):
-            Model = Pool().get(model_name)
-            shipments = set()
-            for line in self.lines:
-                for move in line.moves:
-                    if isinstance(move.shipment, Model):
-                        shipments.add(move.shipment.id)
-            return list(shipments)
-        return method
-
     get_shipments = get_shipments_returns('stock.shipment.out')
     get_shipment_returns = get_shipments_returns('stock.shipment.out.return')
 
-    def search_shipments_returns(model_name):
-        '''
-        Search on shipments or returns
-        '''
-        def method(self, name, clause):
-            nested = clause[0].lstrip(name)
-            if nested:
-                return [('lines.moves.shipment' + nested,)
-                    + tuple(clause[1:3]) + (model_name,) + tuple(clause[3:])]
-            else:
-                if isinstance(clause[2], basestring):
-                    target = 'rec_name'
-                else:
-                    target = 'id'
-                return [('lines.moves.shipment.' + target,)
-                    + tuple(clause[1:3]) + (model_name,)]
-        return classmethod(method)
-
     search_shipments = search_shipments_returns('stock.shipment.out')
     search_shipment_returns = search_shipments_returns(
         'stock.shipment.out.return')
@@ -663,11 +597,14 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             self.raise_user_error('invalid_method', (self.rec_name,))
 
     def get_rec_name(self, name):
+        items = []
         if self.number:
-            return self.number
-        elif self.reference:
-            return '[%s]' % self.reference
-        return '(%s)' % self.id
+            items.append(self.number)
+        if self.reference:
+            items.append('[%s]' % self.reference)
+        if not items:
+            items.append('(%s)' % self.id)
+        return ' '.join(items)
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -684,19 +621,31 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
 
     @classmethod
     def view_attributes(cls):
-        return [('/form//field[@name="comment"]', 'spell', Eval('party_lang'))]
+        attributes = [
+            ('/form//field[@name="comment"]', 'spell', Eval('party_lang'))]
+        if Transaction().context.get('modify_header'):
+            attributes.extend([
+                    ('//group[@id="states"]', 'states', {'invisible': True}),
+                    ('//group[@id="amount_buttons"]',
+                        'states', {'invisible': True}),
+                    ('//page[@name="invoices"]',
+                        'states', {'invisible': True}),
+                    ('//page[@name="shipments"]',
+                        'states', {'invisible': True}),
+                    ])
+        return attributes
 
     @classmethod
     def copy(cls, sales, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['state'] = 'draft'
-        default['number'] = None
-        default['invoice_state'] = 'none'
-        default['invoices_ignored'] = None
-        default['moves'] = None
-        default['shipment_state'] = 'none'
+        else:
+            default = default.copy()
+        default.setdefault('number', None)
+        default.setdefault('invoice_state', 'none')
+        default.setdefault('invoices_ignored', None)
+        default.setdefault('moves', None)
+        default.setdefault('shipment_state', 'none')
         default.setdefault('sale_date', None)
         return super(Sale, cls).copy(sales, default=default)
 
@@ -742,11 +691,10 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     @classmethod
     def store_cache(cls, sales):
         for sale in sales:
-            cls.write([sale], {
-                    'untaxed_amount_cache': sale.untaxed_amount,
-                    'tax_amount_cache': sale.tax_amount,
-                    'total_amount_cache': sale.total_amount,
-                    })
+            sale.untaxed_amount_cache = sale.untaxed_amount
+            sale.tax_amount_cache = sale.tax_amount
+            sale.total_amount_cache = sale.total_amount
+        cls.save(sales)
 
     def _get_invoice_sale(self):
         'Return invoice'
@@ -835,10 +783,10 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         if not moves:
             return
         if shipment_type == 'out':
-            keyfunc = partial(self._group_shipment_key, moves.values())
+            keyfunc = partial(self._group_shipment_key, list(moves.values()))
             Shipment = pool.get('stock.shipment.out')
         elif shipment_type == 'return':
-            keyfunc = partial(self._group_return_key, moves.values())
+            keyfunc = partial(self._group_return_key, list(moves.values()))
             Shipment = pool.get('stock.shipment.out.return')
         moves = moves.items()
         moves = sorted(moves, key=keyfunc)
@@ -895,8 +843,15 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     @ModelView.button
     @Workflow.transition('confirmed')
     def confirm(cls, sales):
+        pool = Pool()
+        Configuration = pool.get('sale.configuration')
         cls.set_sale_date(sales)
         cls.store_cache(sales)
+        config = Configuration(1)
+        with Transaction().set_context(
+                queue_name='sale',
+                queue_scheduled_at=config.sale_process_after):
+            cls.__queue__.process(sales)
 
     @classmethod
     @ModelView.button_action('sale.wizard_invoice_handle_exception')
@@ -918,13 +873,6 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     def do(cls, sales):
         pass
 
-    @classmethod
-    def cron_process(cls):
-        sales = cls.search([
-                ('state', '=', 'confirmed'),
-                ])
-        cls.process(sales)
-
     @classmethod
     @ModelView.button
     def process(cls, sales):
@@ -950,6 +898,11 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         if done:
             cls.do(done)
 
+    @classmethod
+    @ModelView.button_action('sale.wizard_modify_header')
+    def modify_header(cls, sales):
+        pass
+
 
 class SaleIgnoredInvoice(ModelSQL):
     'Sale - Ignored Invoice'
@@ -974,7 +927,6 @@ class SaleRecreatedInvoice(ModelSQL):
 class SaleLine(sequence_ordered(), ModelSQL, ModelView):
     'Sale Line'
     __name__ = 'sale.line'
-    _rec_name = 'description'
     sale = fields.Many2One('sale.sale', 'Sale', ondelete='CASCADE',
         select=True,
         states={
@@ -1048,7 +1000,7 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
                 'invisible': ~Eval('type').in_(['line', 'subtotal']),
                 },
             depends=['type', 'sale_state']), 'get_amount')
-    description = fields.Text('Description', size=None, required=True,
+    description = fields.Text('Description', size=None,
         states={
             'readonly': Eval('sale_state') != 'draft',
             },
@@ -1102,25 +1054,17 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
                     '%(sale)s misses a revenue account.'),
                 'missing_default_account_revenue': ('Sale "%(sale)s" '
                     'misses a default "account revenue".'),
+                'delete_cancel_draft': ('The line "%(line)s" must be on '
+                    'canceled or draft sale to be deleted.'),
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
         super(SaleLine, cls).__register__(module_name)
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
 
-        # Migration from 1.0 comment change into note
-        if table.column_exist('comment'):
-            cursor.execute(*sql_table.update(
-                    columns=[sql_table.note],
-                    values=[sql_table.comment]))
-            table.drop_column('comment', exception=True)
-
-        # Migration from 2.4: drop required on sequence
-        table.not_null_action('sequence', action='remove')
+        # Migration from 4.6: drop required on description
+        table.not_null_action('description', action='remove')
 
     @staticmethod
     def default_type():
@@ -1174,25 +1118,27 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
         '''
         return {}
 
+    @fields.depends('sale', '_parent_sale.currency', '_parent_sale.party',
+        '_parent_sale.sale_date', 'unit', 'product', 'taxes')
     def _get_context_sale_price(self):
         context = {}
-        if getattr(self, 'sale', None):
-            if getattr(self.sale, 'currency', None):
+        if self.sale:
+            if self.sale.currency:
                 context['currency'] = self.sale.currency.id
-            if getattr(self.sale, 'party', None):
+            if self.sale.party:
                 context['customer'] = self.sale.party.id
-            if getattr(self.sale, 'sale_date', None):
-                context['sale_date'] = self.sale.sale_date
+            context['sale_date'] = self.sale.sale_date
         if self.unit:
             context['uom'] = self.unit.id
-        else:
+        elif self.product:
             context['uom'] = self.product.sale_uom.id
-        context['taxes'] = [t.id for t in self.taxes]
+        context['taxes'] = [t.id for t in self.taxes or []]
         return context
 
-    @fields.depends('product', 'unit', 'quantity', 'description', 'sale',
-        '_parent_sale.party', '_parent_sale.currency',
-        '_parent_sale.sale_date')
+    @fields.depends('product', 'unit', 'quantity', 'sale',
+        '_parent_sale.party',
+        methods=['_get_tax_rule_pattern', '_get_context_sale_price',
+        'on_change_with_amount'])
     def on_change_product(self):
         Product = Pool().get('product.product')
 
@@ -1200,11 +1146,8 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
             return
 
         party = None
-        party_context = {}
         if self.sale and self.sale.party:
             party = self.sale.party
-            if party.lang:
-                party_context['language'] = party.lang.code
 
         # Set taxes before unit_price to have taxes in context of sale price
         taxes = []
@@ -1234,10 +1177,6 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
                 self.unit_price = self.unit_price.quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
 
-        if not self.description:
-            with Transaction().set_context(party_context):
-                self.description = Product(self.product.id).rec_name
-
         self.type = 'line'
         self.amount = self.on_change_with_amount()
 
@@ -1246,9 +1185,8 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
         if self.product:
             return self.product.default_uom_category.id
 
-    @fields.depends('product', 'quantity', 'unit', 'taxes', 'sale',
-        '_parent_sale.currency', '_parent_sale.party',
-        '_parent_sale.sale_date')
+    @fields.depends('product', 'quantity',
+        methods=['_get_context_sale_price'])
     def on_change_quantity(self):
         Product = Pool().get('product.product')
 
@@ -1263,11 +1201,11 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
                 self.unit_price = self.unit_price.quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
 
-    @fields.depends(methods=['quantity'])
+    @fields.depends(methods=['on_change_quantity'])
     def on_change_unit(self):
         self.on_change_quantity()
 
-    @fields.depends(methods=['quantity'])
+    @fields.depends(methods=['on_change_quantity'])
     def on_change_taxes(self):
         self.on_change_quantity()
 
@@ -1321,19 +1259,20 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
     @fields.depends('product', 'quantity', 'moves', 'sale',
         '_parent_sale.sale_date')
     def on_change_with_shipping_date(self, name=None):
+        pool = Pool()
+        Date = pool.get('ir.date')
         if self.moves:
             dates = filter(
                 None, (m.effective_date or m.planned_date for m in self.moves
                     if m.state != 'cancel'))
-            if dates:
-                return min(dates)
-            else:
-                return
+            return min(dates, default=None)
         if self.product and self.quantity is not None and self.quantity > 0:
             date = self.sale.sale_date if self.sale else None
             shipping_date = self.product.compute_shipping_date(date=date)
             if shipping_date == datetime.date.max:
-                return None
+                shipping_date = None
+            if shipping_date:
+                shipping_date = max(shipping_date, Date.today())
             return shipping_date
 
     @fields.depends('sale', '_parent_sale.state')
@@ -1382,11 +1321,8 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
                         'product': self.product.rec_name,
                         })
         else:
-            for name in ['default_product_account_revenue',
-                    'default_category_account_revenue']:
-                invoice_line.account = account_config.get_multivalue(name)
-                if invoice_line.account:
-                    break
+            invoice_line.account = account_config.get_multivalue(
+                'default_category_account_revenue')
             if not invoice_line.account:
                 self.raise_user_error('missing_default_account_revenue', {
                         'sale': self.sale.rec_name,
@@ -1537,21 +1473,44 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
                         invoice_lines.append(invoice_line)
         return invoice_lines
 
+    def get_rec_name(self, name):
+        if self.product:
+            return '%s @ %s' % (self.product.rec_name, self.sale.rec_name)
+        else:
+            return self.sale.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return ['OR',
+            ('sale.rec_name',) + tuple(clause[1:]),
+            ('product.rec_name',) + tuple(clause[1:]),
+            ]
+
     @classmethod
     def view_attributes(cls):
         return [
             ('/form//field[@name="note"]|/form//field[@name="description"]',
                 'spell', Eval('_parent_sale', {}).get('party_lang'))]
 
+    @classmethod
+    def delete(cls, lines):
+        for line in lines:
+            if line.sale_state not in {'cancel', 'draft'}:
+                cls.raise_user_error('delete_cancel_draft', {
+                        'line': line.rec_name,
+                        })
+        super(SaleLine, cls).delete(lines)
+
     @classmethod
     def copy(cls, lines, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['moves'] = None
-        default['moves_ignored'] = None
-        default['moves_recreated'] = None
-        default['invoice_lines'] = None
+        else:
+            default = default.copy()
+        default.setdefault('moves', None)
+        default.setdefault('moves_ignored', None)
+        default.setdefault('moves_recreated', None)
+        default.setdefault('invoice_lines', None)
         return super(SaleLine, cls).copy(lines, default=default)
 
 
@@ -1593,6 +1552,14 @@ class SaleReport(CompanyReport):
         with Transaction().set_context(address_with_party=True):
             return super(SaleReport, cls).execute(ids, data)
 
+    @classmethod
+    def get_context(cls, records, data):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        context = super(SaleReport, cls).get_context(records, data)
+        context['today'] = Date.today()
+        return context
+
 
 class OpenCustomer(Wizard):
     'Open Customers'
@@ -1629,20 +1596,6 @@ class HandleShipmentExceptionAsk(ModelView):
     domain_moves = fields.Many2Many(
         'stock.move', None, None, 'Domain Moves')
 
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        model = Table('ir_model')
-        # Migration from 1.2: packing renamed into shipment
-        cursor.execute(*model.update(
-                columns=[model.model],
-                values=[Overlay(model.model, 'shipment',
-                        Position('packing', model.model),
-                        len('packing'))],
-                where=model.model.like('%packing%')
-                & (model.module == module_name)))
-        super(HandleShipmentExceptionAsk, cls).__register__(module_name)
-
 
 class HandleShipmentException(Wizard):
     'Handle Shipment Exception'
@@ -1695,7 +1648,7 @@ class HandleShipmentException(Wizard):
                     'moves_ignored': [('add', moves_ignored)],
                     'moves_recreated': [('add', moves_recreated)],
                     })
-        Sale.process([sale])
+        Sale.__queue__.process([sale])
         return 'end'
 
 
@@ -1755,7 +1708,7 @@ class HandleInvoiceException(Wizard):
                 'invoices_ignored': [('add', invoices_ignored)],
                 'invoices_recreated': [('add', invoices_recreated)],
                 })
-        Sale.process([sale])
+        Sale.__queue__.process([sale])
         return 'end'
 
 
@@ -1791,3 +1744,73 @@ class ReturnSale(Wizard):
         if len(return_sales) == 1:
             action['views'].reverse()
         return action, data
+
+
+class ModifyHeaderStateView(StateView):
+    def get_view(self, wizard, state_name):
+        with Transaction().set_context(modify_header=True):
+            return super(ModifyHeaderStateView, self).get_view(
+                wizard, state_name)
+
+
+class ModifyHeader(Wizard):
+    "Modify Header"
+    __name__ = 'sale.modify_header'
+    start = ModifyHeaderStateView('sale.sale',
+        'sale.modify_header_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Modify", 'modify', 'tryton-ok', default=True),
+            ])
+    modify = StateTransition()
+
+    @classmethod
+    def __setup__(cls):
+        super(ModifyHeader, cls).__setup__()
+        cls._error_messages.update({
+                'not_in_draft': (
+                    'The sale "%(sale)s" must be in draft to modify header.'),
+                })
+
+    def get_sale(self):
+        pool = Pool()
+        Sale = pool.get('sale.sale')
+
+        sale = Sale(Transaction().context['active_id'])
+        if sale.state != 'draft':
+            self.raise_user_error('not_in_draft', {
+                    'sale': sale.rec_name,
+                    })
+        return sale
+
+    def default_start(self, fields):
+        sale = self.get_sale()
+        defaults = {}
+        for fieldname in fields:
+            value = getattr(sale, fieldname)
+            if isinstance(value, Model):
+                if getattr(sale.__class__, fieldname)._type == 'reference':
+                    value = str(value)
+                else:
+                    value = value.id
+            elif isinstance(value, (list, tuple)):
+                value = [r.id for r in value]
+            defaults[fieldname] = value
+
+        # Mimic an empty sale in draft state to get the fields' states right
+        defaults['lines'] = []
+        return defaults
+
+    def transition_modify(self):
+        pool = Pool()
+        Line = pool.get('sale.line')
+
+        sale = self.get_sale()
+        sale.__class__.write([sale], self.start._save_values)
+
+        # Call on_change after the save to ensure parent sale
+        # has the modified values
+        for line in sale.lines:
+            line.on_change_product()
+        Line.save(sale.lines)
+
+        return 'end'
